# Import error specific framework elements under errors
import cfw.framework.errors as errors

# Import the framework discovery members into our toplevel scope
from cfw.framework.discovery import command, dispatch

# Import argument specific framework elements into our toplevel scope
from cfw.framework.args import Argument, WildcardArgument, ListArgument, Flag, Positional, ListPositional, \
    WildcardPositional



from cfw.framework.util import GREEDY_WHITESPACE_RE

# Formatting guides
_FIRST_COLUMN_LEN = 30
_SECOND_COLUMN_LEN = 56

# 32 - 4 (spacing) - 2 (leading space)
_MAX_FIRST_COLUMN_LEN = 26
_MAX_LINE_LEN = 88
_MAX_SECOND_COLUMN_LINES = 4


def _sanitize(input_value):
    sanitize_me = input_value
    if not isinstance(input_value, str):
        sanitize_me = str(input_value)

    return sanitize_me.strip().replace('\n', ' ')


def _sanitize_split(input):
    return GREEDY_WHITESPACE_RE.split(_sanitize(input))


def format_one_column_output(first):
    first_output = ''

    buf = ''
    for word in _sanitize_split(first):
        if len(buf) + len(word) > _MAX_LINE_LEN:
            first_output += '{}\n'.format(buf)
            buf = word

        elif len(buf) == 0:
            buf = word

        else:
            buf += ' {}'.format(word)

    if len(buf) > 0:
        first_output += buf

    return first_output


def format_two_column_output(first, second):
    first_output = '  {}'.format(_sanitize(first))
    if len(first_output) > _MAX_FIRST_COLUMN_LEN:
        first_output = '  {}...'.format(first[:_MAX_FIRST_COLUMN_LEN - 3])

    add_nl_buffer = False
    second_output = ''.join([' ' for _ in range(_FIRST_COLUMN_LEN - len(first_output))])
    if second is not None:
        buf = ''
        lines = 1
        for word in _sanitize_split(second):
            if len(buf) + len(word) > _SECOND_COLUMN_LEN:
                # Make sure to output an additional newline for this output since we broke lines
                add_nl_buffer = True

                # Append to our output buffer and set our working buffer to the word that wouldn't fit
                second_output += '{}\n{}'.format(buf, ''.join([' ' for _ in range(_FIRST_COLUMN_LEN)]))
                buf = word

                # Increase the number of lines - if there are too many break here.
                lines += 1
                if lines >= _MAX_SECOND_COLUMN_LINES:
                    buf = '{}...'.format(buf)
                    break

            elif len(buf) == 0:
                buf = word

            else:
                buf += ' {}'.format(word)

        if len(buf) > 0:
            second_output += buf

    return '{}{}{}'.format(first_output, second_output, '' if add_nl_buffer is False else '\n')



"""
The function defintions below should create a tree structure like the following:

root
  |- first
  |- test
       |- second
"""

import cfw


# Function and definition for the first command
@cfw.command(arguments=[
    cfw.Flag('-v', '--verbose', help='Run with more output.'),
    cfw.Argument('-r', '--default', help='This is a default argument.'),
    cfw.Positional(name='positional_arg', help='This is a required positional argument.'),
    cfw.Argument('-o', '--optional', help='This is an optional argument.')
])
def first(verbose, required, positional, optional=None):
    """
    Lorem ipsum dolor sit amet, consectetur adipiscing elit. Donec auctor turpis a ligula sollicitudin pellentesque.
    Maecenas quis feugiat neque. Vestibulum eu sem id augue iaculis elementum eu vel dolor. Suspendisse aliquet orci
    nec ipsum dapibus, non maximus felis pretium. In auctor eleifend neque, quis bibendum turpis convallis ac.
    Suspendisse dignissim auctor leo, vitae mollis dolor aliquam id. Nunc id leo placerat, commodo tellus sed,
    fringilla lectus. Cras posuere condimentum urna a volutpat. Vestibulum convallis interdum euismod. Duis dapibus
    sagittis erat, a iaculis urna porta nec. Vivamus pulvinar dolor lectus, et tristique tellus efficitur vel.
    Orci varius natoque penatibus et magnis dis parturient montes, nascetur ridiculus mus. Mauris porttitor nunc
    a pellentesque semper.
    """
    print('{} {}'.format(verbose, required))


# Help output may be placed in the command definition if you wish to be explicit or rely on docstrings
# for other features and or documentation.
#
# By specifying a path for this command we're explicitly telling CFW that this is part of a nested command structure
# and that this command belongs with the descendants of the test command. If there is no corresponding test command
# defined elsewhere, then CFW will create a simple stub that handles dispatch and collates documentation for you.
@cfw.command(path='test', help='A nested command that lets you test nested paths.')
def second():
    pass


# Testcase for deeply nested commands that expect their path to be simple command aggregators
@cfw.command(path='command path')
def nested():
    pass


def main():
    cfw.dispatch('cfw.testapp', help='This is a test CLI for CFW')


if __name__ == '__main__':
    main()



class CommandError(Exception):
    def __init__(self, msg=''):
        super(CommandError, self).__init__()

        self.msg = msg

    def __str__(self):
        return self.msg


class CommandArgumentError(CommandError):
    pass


class CommandNotFoundError(CommandError):
    pass


class CommandDependencyError(CommandError):
    def __str__(self):
        return 'Command Dependency Error: {}'.format(self.msg)



import importlib
import os
import sys

from cfw.framework.command import CommandWrapper, CommandTrie
from cfw.framework.errors import CommandDependencyError, CommandError

_PYTHON_SRC_CODE_EXT = '.py'
_PYTHON_BYTE_CODE_EXT = '.pyc'
_PRIVATE_NAME_PREFIX = '__'
_IGNORE_LIST = ('__pycache__', '__init__.py',)
_PYTHON_EXTENSIONS = (_PYTHON_SRC_CODE_EXT, _PYTHON_BYTE_CODE_EXT,)
_PYTHON_MODULE_INIT_FILE = '__init__.py'


def _is_python_src_file(filename):
    return filename.endswith(_PYTHON_SRC_CODE_EXT)


def _is_relative(module):
    return module.startswith('.')


def _file_extension(path):
    ext_parts = os.path.splitext(path)
    return ext_parts[1] if len(ext_parts) > 1 else ''


def command(*args, **kwargs):
    """
    Just a marker annotation for code documentation for now. Might work to auto-pull in
    command line functions but that's for later.

    :return:
    """
    func = None
    invoked = bool(not args or kwargs)
    if invoked is False:
        func, args = args[0], ()

    def _factory(target_func):
        return CommandWrapper(target_func, *args, **kwargs)

    return _factory if invoked is True else _factory(func)


def dispatch(module, package=None, argv=None, verbose=False, help=None):
    if _is_relative(module) is True and package is None:
        raise CommandError('Attempting an import of a relative module requires the package argument to be specified.')

    if argv is None:
        argv = sys.argv

    trie = scan(argv[0], module, package, verbose, help)
    trie.dispatch(argv)


def scan(cli_call_name, module, package, verbose, help):
    """
    This crawls all of the modules below us and imports them recursively
    :return:
    """
    root_module = importlib.import_module(module, package=package)
    root_path = root_module.__file__

    if verbose:
        print('Scanning module {} starting at file path: {}'.format(module, root_path))

    # Format the module name correctly if this is a relative import we're starting with
    target_module = module
    if _is_relative(target_module):
        target_module = package

    # Search path changes if the __file__ entry is a python init file and not a directory
    search_path = root_path
    if search_path.endswith(_PYTHON_MODULE_INIT_FILE):
        search_path = os.path.dirname(root_path)

    # First identify all submodules
    submodule_names = list()

    # If our search path is not a directory, move on
    if os.path.isdir(search_path):
        for filename in os.listdir(search_path):
            if filename in _IGNORE_LIST:
                continue

            abs_path = os.path.join(search_path, filename)
            init_path = os.path.join(abs_path, _PYTHON_MODULE_INIT_FILE)

            module_name = ''
            if os.path.isdir(abs_path) and os.path.exists(init_path):
                # Figure out if we're dealing with a directory that has the init file
                module_name = '.'.join((target_module, filename,))
            elif _is_python_src_file(filename):
                # Is it a python source file that's stand-alone?
                file_module_name = os.path.splitext(filename)[0]
                module_name = '.'.join((target_module, file_module_name,))
            else:
                # I don't like this continue but avoiding the print statement twice is a nice consequence
                continue

            if verbose:
                print('Adding module {} to the scan list.'.format(module_name))

            # Add the module to our scan and import list
            submodule_names.append(module_name)

    # Load the modules
    submodules = [importlib.import_module(n) for n in submodule_names]

    # Add the root module since that's part of the scan
    submodules.append(root_module)

    # Load and scan the submodules for command components
    command_components = list()
    for submodule in submodules:
        for component_name in dir(submodule):
            component = getattr(submodule, component_name)
            if isinstance(component, CommandWrapper):
                if verbose:
                    print('Found command component: {}'.format(component))

                command_components.append(component)

    # Build our command trie with collected components and perform rudimentary
    # dependency resolution for command paths
    command_trie = CommandTrie(cli_call_name, help=help)
    while len(command_components) > 0:
        delete_list = list()
        for idx in range(0, len(command_components)):
            command = command_components[idx]

            if command_trie.insert(command) is True:
                if verbose:
                    print('Inserted {}'.format(command))

                delete_list.append(idx)
                break

        if len(delete_list) == 0:
            raise CommandDependencyError('Dependency resolution error!')

        for idx in reversed(sorted(delete_list)):
            command_components.pop(idx)

    return command_trie



from cfw.framework.errors import CommandArgumentError


class ArgumentDefinition(object):
    def __init__(self, short_form=None, long_form=None, name=None, help=None):
        self.short_form = short_form
        self.long_form = long_form
        self.name = name
        self.help = help
        self.keyword = None
        self.required = True
        self.default = None
        self.has_default = False

    @property
    def positional(self):
        return False

    def set_default(self, value):
        self.default = value
        self.has_default = True

    def check(self):
        if self.short_form is None and self.long_form is None:
            raise CommandArgumentError('No valid CLI form specified for argument: {}'.format(self.keyword))

    def forms(self):
        """
        Returns a formatted string with the matchable forms for the argument
        :return:
        """
        forms = ''
        if self.short_form is not None:
            forms += self.short_form

        if self.long_form is not None:
            if len(forms) > 0:
                forms += ', '

            forms += self.long_form

        return forms

    def matches(self, arg):
        matches = self.short_form is not None and arg == self.short_form

        if matches is False:
            matches = self.long_form is not None and arg == self.long_form

        return matches

    def __str__(self):
        if self.name is not None:
            return self.name
        return self.forms()


class Argument(ArgumentDefinition):
    pass


class Positional(ArgumentDefinition):
    @property
    def positional(self):
        return True

    def check(self):
        if self.name is None:
            raise CommandArgumentError('Positional arguments require a name.')

    def matches(self, arg):
        return True


class ListPositional(Positional):
    pass


class WildcardPositional(Positional):
    pass


class WildcardArgument(ArgumentDefinition):
    pass


class ListArgument(ArgumentDefinition):
    pass


class Flag(ArgumentDefinition):
    pass



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name='cfw',
    version='0.1.11',
    url='https://github.com/zinic/python-cfw',
    description='Python Command Line Interface Library',

    author='John Hopper',
    author_email='john.hopper@jpserver.net',

    packages=find_packages(),

    license='MIT'
)



import inspect
import itertools
import sys

from collections import defaultdict

from cfw.framework.args import *
from cfw.framework.formatting import format_two_column_output, format_one_column_output
from cfw.framework.errors import CommandError, CommandArgumentError
from cfw.framework.util import GREEDY_WHITESPACE_RE

_EXEC_OK = 'OK'
_PRINT_HELP = 'PRINT_HELP'

_PRIVATE_NAME_PREFIX = '__'
_HELP_ARGUMENTS = ('-h', '-?', '--help',)


def command_stub(name, path=None, help=''):
    def _stub_closure():
        return _PRINT_HELP

    return CommandWrapper(_stub_closure, name=name, path=path, help=help)


class CommandTrie(object):
    def __init__(self, cli_call_name, help=None):
        # This name is passed to us so that our output is accurate according to what the user typed in or
        # calls the program we're processing arguments for.
        self.cli_call_name = cli_call_name

        # Root commands are nameless
        self._root_cmd = CommandNode(command_stub('', help=help), is_root=True)

    def dispatch(self, argv=None):
        """
        Attempts to dispatch to the correct command function given a set of argdefs. If no argdefs are
        specified then sys.argv is used instead.

        :param argv:
        :return: True if the command dispatched successfully, False otherwise.
        """

        # If we weren't given any argdefs simply assume the user wanted to operate on sys.argv
        if argv is None:
            argv = sys.argv

        # No argdefs usually means the user is hunting for the help output, otherwise
        # if the first argument is a help argument, print help as well.
        if len(argv) <= 1 or argv[1] in _HELP_ARGUMENTS:
            self.print_help()
            return False

        # List for capturing anything that looks like a CLI argument and not part of the command path
        args_list = list()

        # Always start at argv[1] for the command path
        cursor = self._root_cmd
        for arg in argv[1:]:
            # Attempt to match the argument against our path - anything that doesn't match
            # gets added to the argument list
            potential_next = cursor.get(arg)
            if potential_next is not None:
                cursor = potential_next
            else:
                args_list.append(arg)

        # If the cursor is still pointing at our root dictionary then no command parts matched
        if cursor is self._root_cmd:
            print('Unknown command: {}\n'.format(' '.join(argv)))

            self.print_help()
            return False

        cursor.exec(self.cli_call_name, args_list)
        return True

    def print_help(self):
        self._root_cmd.print_help(self.cli_call_name)

    def find(self, path):
        # Find the root CN first
        root_path = path[0]

        # Track what we matched
        matched_path = [root_path]

        cursor = self._root_cmd.get(root_path)
        if cursor is not None:
            # Look up descendants till we exhaust the path
            for path_part in path[1:]:
                # Track where we are
                matched_path.append(path_part)

                # Check to see if the next part has a matching command
                cursor = cursor.get(path_part)

                # If there's no matching command at this part of the path, return where we
                # left off in the path
                if cursor is None:
                    break

        return cursor, matched_path

    def insert(self, command):
        if command.depth == 0:
            self._root_cmd.append(CommandNode(command))

        else:
            target, last_path = self.find(command.path)
            while target is None:
                # Get the name of the command we choked on
                missing_cmd = last_path[len(last_path) - 1]

                # Recreate the path string for insertion
                path_str = None
                if len(last_path) > 1:
                    path_str = ' '.join(last_path[:len(last_path) - 1])

                # Create a command stub - this may be overwritten by further, more specific inserts
                self.insert(command_stub(missing_cmd, path=path_str))

                # Attempt to resolve our path again
                target, last_path = self.find(command.path)

            target.append(CommandNode(command))

        return True


class CommandNode(object):
    def __init__(self, cmd, is_root=False):
        self.cmd = cmd
        self.is_root = is_root
        self.descendants = dict()

    def append(self, cmd):
        self.descendants[cmd.name] = cmd

    def get(self, name):
        return self.descendants.get(name)

    def print_help(self, cli_call_name):
        cmd_path = '{} {}'.format(' '.join(self.cmd.path), self.name) if self.cmd.depth > 0 else self.name
        cmd_spec = 'Usage: {} {}'.format(cli_call_name, cmd_path)

        # Building an arg spec string is important for things like positional arguments where the order
        # of the argument determines where it is slotted in the target function
        cmd_arg_spec = ''
        for argdef in self.cmd.argdefs:
            if argdef.positional is True:
                # If this isn't the first positional arg added to the cmd_arg_spec then we need as space
                if len(cmd_arg_spec) > 0:
                    cmd_arg_spec += ' '

                cmd_arg_spec += '<{}>'.format(argdef.name)

        # If there were positional arguments we need to append the cmd_arg_spec to the cmd_spec for output
        if len(cmd_arg_spec) > 0:
            cmd_spec += ' {}'.format(cmd_arg_spec)

        # Usage line first with the name we were called by on the CLI
        print('{}\n'.format(cmd_spec))

        # Only output the help info if we have a help specified that is of non-zero length
        if self.help is not None and len(self.help) > 0:
            print('{}\n'.format(format_one_column_output(self.help)))

        # List subcommands if any
        if len(self.descendants) > 0:
            self._print_subcmd_help()

            # Extra newline
            print('')

        # Any arguments from the command it self should be printed now
        self.cmd.print_help()

    def _print_subcmd_help(self):
        output = 'Available Commands:\n'

        for name, subcmd in self.descendants.items():
            output += '{}\n'.format(format_two_column_output(name, subcmd.help))

        print(output)

    @property
    def help(self):
        return self.cmd.help if self.cmd.help is not None else 'No help output specified.'

    @property
    def name(self):
        return self.cmd.name

    def exec(self, cli_call_name, argv):
        result = self.cmd(argv)

        if result == _PRINT_HELP:
            self.print_help(cli_call_name)

    def __str__(self):
        return 'CommandNode({})'.format(self.name)


class ArgumentIterator(object):
    def __init__(self, argv):
        self._idx = 0
        self._argv = argv

    def finish(self):
        self._idx = len(self._argv)

    def advance(self, steps=1):
        self._idx += steps

    def get(self):
        return self._argv[self._idx]

    def get_rest(self):
        return self._argv[self._idx:]

    @property
    def on_last(self):
        return self._idx + 1 == len(self._argv)

    @property
    def empty(self):
        return self._idx >= len(self._argv)


class ArgumentMapper(object):
    def __init__(self, arg_defs):
        # Track different argument types to make searching more deterministic
        self.positionals = list()
        self.non_positionals = list()

        for argdef in arg_defs:
            if argdef.positional is True:
                self.positionals.append(argdef)
            else:
                self.non_positionals.append(argdef)

    def _match_arg(self, arg):
        # Search non-positional argument definitions first
        for argdef in self.non_positionals:
            if argdef.matches(arg):
                return argdef
        return None

    def _prepare_kwargs(self):
        kwargs = defaultdict(list)

        # Map all flags first as False
        for arg_def in itertools.chain(self.positionals, self.non_positionals):
            if arg_def.default is not None:
                # Function argument defaults beat out our typing
                kwargs[arg_def.keyword] = arg_def.default

        return kwargs

    def map_to_kwargs(self, argv):
        arg_source = ArgumentIterator(argv)
        kwargs = self._prepare_kwargs()

        while arg_source.empty is False:
            # Get the next argument
            arg = arg_source.get()

            # Try to match the arg against non-positional argdefs first
            argdef = self._match_arg(arg)
            if argdef is None:
                # If there are no positional arguments remaining then this argument is unknown
                if len(self.positionals) == 0:
                    raise CommandError('Unknown argument: {}'.format(arg))

                # Select the first positional argument
                argdef = self.positionals[0]

            # Attempt to gather up the value that's represented by the argument
            if isinstance(argdef, Flag):
                kwargs[argdef.keyword] = True
                arg_source.advance()

            elif isinstance(argdef, ListArgument):
                arg_source.advance()
                kwargs[argdef.keyword].append(arg_source.get())
                arg_source.advance()

            elif isinstance(argdef, WildcardArgument):
                arg_source.advance()
                kwargs[argdef.keyword].extend(arg_source.get_rest())
                arg_source.finish()

            elif isinstance(argdef, Argument):
                arg_source.advance()
                kwargs[argdef.keyword] = arg_source.get()
                arg_source.advance()

            elif isinstance(argdef, ListPositional):
                # First remove this positional argdef from our list of positional arg defs
                self.positionals.pop(0)

                # Add the arg as our value
                kwargs[argdef.keyword].append(arg_source.get())
                arg_source.advance()

                # Continue consuming arguments until the next match or until we reach a point
                # where other positional arguments expect to be filled in
                while not arg_source.on_last and not arg_source.empty:
                    # If an argument definition matches then we're done with this list
                    if self._match_arg(arg_source.get()) is not None:
                        break

                    arg_source.advance()

            elif isinstance(argdef, WildcardPositional):
                # First remove this positional argdef from our list of positional arg defs
                self.positionals.pop(0)

                kwargs[argdef.keyword].extend(arg_source.get_rest())
                arg_source.finish()

            elif isinstance(argdef, Positional):
                # First remove this positional argdef from our list of positional arg defs
                self.positionals.pop(0)

                kwargs[argdef.keyword] = arg_source.get()
                arg_source.advance()

        return kwargs


class CommandWrapper(object):
    def __init__(self, func, name=None, path=None, help=None, arguments=None):
        self.func = func
        self.name = name
        self.path = list()
        self.path_spec = path
        self.help = help
        self.argdefs = arguments

        self._func_argspec = inspect.getfullargspec(self.func)

        if self.name is None:
            # If there's no name specified then use the name of the fuction
            # instead
            self.name = self.func.__name__

        if self.path_spec is not None:
            # If there's a path_spec, parse it into our path array
            self.path.extend(GREEDY_WHITESPACE_RE.split(self.path_spec))

        if self.help is None and self.func.__doc__ is not None:
            # If there's a valid docstring for the function but no help output
            # provided then use the docstring instead
            self.help = self.func.__doc__

        if self.argdefs is None:
            # Make sure that self.argdefs is never None
            self.argdefs = list()

        # Process our definitions and do some sanity checks
        self._process_arg_defs()

    def _process_arg_defs(self):
        # Iterator to track our position in the arg keyword list in reverse
        arg_kw_iter = reversed(self._func_argspec.args)

        # We create the arg_defaults iterator with an empty list first to avoid any weird
        # special-case checks while popping off default values
        arg_default_iter = iter(list())
        if self._func_argspec.defaults is not None:
            # If there is a set of defaults in the argspec, iterate through them in reverse
            arg_default_iter = reversed(self._func_argspec.defaults)

        for arg_def in reversed(self.argdefs):
            # Check the annotation for hygiene first
            if arg_def.short_form in _HELP_ARGUMENTS or arg_def.long_form in _HELP_ARGUMENTS:
                raise CommandArgumentError('Arguments may not carry the signature of: {}'.format(_HELP_ARGUMENTS))

            # Assign the matching function keyword
            try:
                arg_def.keyword = next(arg_kw_iter)
            except StopIteration:
                raise CommandArgumentError('CLI argument {} defined but function {} has no answering argument.'.format(
                    arg_def, self.func.__name__)) from None

            # Map the default if any
            next_default = next(arg_default_iter, None)
            if next_default is not None:
                arg_def.set_default(next_default)

            # If there's no default but the argument is a flag, default to False
            if arg_def.has_default is False and isinstance(arg_def, Flag):
                arg_def.set_default(False)

        # Run sanity checks now that the argument definitions have been filled out with the remainder of
        # important details
        for arg_def in self.argdefs:
            arg_def.check()

    def print_help(self):
        # If there aren't any args, tell the user
        if len(self.argdefs) == 0:
            print('This command has no arguments specified.')
            return

        # Try to print out detailed argument help
        print('Arguments:')
        for argdef in self.argdefs:
            print(format_two_column_output(argdef, argdef.help))

    @property
    def depth(self):
        return len(self.path)

    def __call__(self, argv):
        # Scan argdefs for potential help requests
        for arg in argv:
            if arg in _HELP_ARGUMENTS:
                return _PRINT_HELP

        # Generate a kwargs dict
        kwargs = ArgumentMapper(self.argdefs).map_to_kwargs(argv)

        # Last but not least, we test to make sure all required arguments are provided
        for arg_def in self.argdefs:
            if arg_def.has_default is True and kwargs.get(arg_def.keyword) is None:
                print('Missing required argument: {}\n'.format(arg_def.short_form))
                return _PRINT_HELP

        # Hand off the args to the real receiver
        result = self.func(**kwargs)
        if result is _PRINT_HELP:
            return result

        return _EXEC_OK

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name
